/*
 * Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */
#include <rte_ethdev.h>

#include <doca_log.h>

#include <pack.h>
#include <utils.h>

#include "doca_flow.h"
#include "flow_encrypt.h"

DOCA_LOG_REGISTER(IPSEC_SECURITY_GW::flow_encrypt);

#define ENCAP_DST_IP_IDX_IP4 30		  /* index in encap raw data for destination IPv4 */
#define ENCAP_DST_IP_IDX_IP6 38		  /* index in encap raw data for destination IPv4 */
#define ENCAP_IP_ID_IDX_IP4 18		  /* index in encap raw data for IPv4 ID */
#define ENCAP_DST_UDP_PORT_IDX 2	  /* index in encap raw data for UDP destination port */
#define ENCAP_ESP_SPI_IDX_TUNNEL_IP4 34	  /* index in encap raw data for esp SPI in IPv4 tunnel */
#define ENCAP_ESP_SPI_IDX_TUNNEL_IP6 54	  /* index in encap raw data for esp SPI in IPv6 tunnel */
#define ENCAP_ESP_SPI_IDX_TRANSPORT 0	  /* index in encap raw data for esp SPI in transport mode*/
#define ENCAP_ESP_SPI_IDX_UDP_TRANSPORT 8 /* index in encap raw data for esp SPI in transport over UDP mode*/
#define ENCAP_ESP_SN_IDX_TUNNEL_IP4 38	  /* index in encap raw data for esp SN in IPv4 tunnel */
#define ENCAP_ESP_SN_IDX_TUNNEL_IP6 58	  /* index in encap raw data for esp SN in IPv6 tunnel */
#define ENCAP_ESP_SN_IDX_TRANSPORT 4	  /* index in encap raw data for esp SN in transport mode*/
#define ENCAP_ESP_SN_IDX_UDP_TRANSPORT 12 /* index in encap raw data for esp SN in transport over UDP mode*/

#define PADDING_ALIGN 4 /* padding alignment */

static const uint8_t esp_pad_bytes[15] = {
	1,
	2,
	3,
	4,
	5,
	6,
	7,
	8,
	9,
	10,
	11,
	12,
	13,
	14,
	15,
};

static uint16_t current_ip_id; /* Incremented for each new packet */

/*
 * Create reformat data for encapsulation in transport mode, and copy it to reformat_data pointer
 *
 * @rule [in]: current rule for encapsulation
 * @sw_sn_inc [in]: if true, sequence number will be incremented in software
 * @reformat_data [out]: pointer to created data
 * @reformat_data_sz [out]: data size
 */
static void create_transport_encap(struct encrypt_rule *rule,
				   bool sw_sn_inc,
				   uint8_t *reformat_data,
				   uint16_t *reformat_data_sz)
{
	uint8_t reformat_encap_data[16] = {
		0x00,
		0x00,
		0x00,
		0x00, /* SPI */
		0x00,
		0x00,
		0x00,
		0x00, /* SN */
		0x00,
		0x00,
		0x00,
		0x00, /* IV */
		0x00,
		0x00,
		0x00,
		0x00,
	};

	reformat_encap_data[ENCAP_ESP_SPI_IDX_TRANSPORT] = GET_BYTE(rule->esp_spi, 3);
	reformat_encap_data[ENCAP_ESP_SPI_IDX_TRANSPORT + 1] = GET_BYTE(rule->esp_spi, 2);
	reformat_encap_data[ENCAP_ESP_SPI_IDX_TRANSPORT + 2] = GET_BYTE(rule->esp_spi, 1);
	reformat_encap_data[ENCAP_ESP_SPI_IDX_TRANSPORT + 3] = GET_BYTE(rule->esp_spi, 0);

	if (sw_sn_inc == true) {
		reformat_encap_data[ENCAP_ESP_SN_IDX_TRANSPORT] = GET_BYTE(rule->current_sn, 3);
		reformat_encap_data[ENCAP_ESP_SN_IDX_TRANSPORT + 1] = GET_BYTE(rule->current_sn, 2);
		reformat_encap_data[ENCAP_ESP_SN_IDX_TRANSPORT + 2] = GET_BYTE(rule->current_sn, 1);
		reformat_encap_data[ENCAP_ESP_SN_IDX_TRANSPORT + 3] = GET_BYTE(rule->current_sn, 0);
	}

	memcpy(reformat_data, reformat_encap_data, sizeof(reformat_encap_data));
	*reformat_data_sz = sizeof(reformat_encap_data);
}

/*
 * Create reformat data for encapsulation in UDP transport mode, and copy it to reformat_data pointer
 *
 * @rule [in]: current rule for encapsulation
 * @sw_sn_inc [in]: if true, sequence number will be incremented in software
 * @reformat_data [out]: pointer to created data
 * @reformat_data_sz [out]: data size
 */
static void create_udp_transport_encap(struct encrypt_rule *rule,
				       bool sw_sn_inc,
				       uint8_t *reformat_data,
				       uint16_t *reformat_data_sz)
{
	uint16_t udp_dst_port = 4500;
	uint8_t reformat_encap_data[24] = {
		0x30, 0x39, 0x00, 0x00, /* UDP src/dst */
		0x00, 0xa4, 0x00, 0x00, /* USD sum/len */
		0x00, 0x00, 0x00, 0x00, /* SPI */
		0x00, 0x00, 0x00, 0x00, /* SN */
		0x00, 0x00, 0x00, 0x00, /* IV */
		0x00, 0x00, 0x00, 0x00,
	};

	reformat_encap_data[ENCAP_ESP_SPI_IDX_UDP_TRANSPORT] = GET_BYTE(rule->esp_spi, 3);
	reformat_encap_data[ENCAP_ESP_SPI_IDX_UDP_TRANSPORT + 1] = GET_BYTE(rule->esp_spi, 2);
	reformat_encap_data[ENCAP_ESP_SPI_IDX_UDP_TRANSPORT + 2] = GET_BYTE(rule->esp_spi, 1);
	reformat_encap_data[ENCAP_ESP_SPI_IDX_UDP_TRANSPORT + 3] = GET_BYTE(rule->esp_spi, 0);

	reformat_encap_data[ENCAP_DST_UDP_PORT_IDX] = GET_BYTE(udp_dst_port, 1);
	reformat_encap_data[ENCAP_DST_UDP_PORT_IDX + 1] = GET_BYTE(udp_dst_port, 0);

	if (sw_sn_inc == true) {
		reformat_encap_data[ENCAP_ESP_SN_IDX_UDP_TRANSPORT] = GET_BYTE(rule->current_sn, 3);
		reformat_encap_data[ENCAP_ESP_SN_IDX_UDP_TRANSPORT + 1] = GET_BYTE(rule->current_sn, 2);
		reformat_encap_data[ENCAP_ESP_SN_IDX_UDP_TRANSPORT + 2] = GET_BYTE(rule->current_sn, 1);
		reformat_encap_data[ENCAP_ESP_SN_IDX_UDP_TRANSPORT + 3] = GET_BYTE(rule->current_sn, 0);
	}

	memcpy(reformat_data, reformat_encap_data, sizeof(reformat_encap_data));
	*reformat_data_sz = sizeof(reformat_encap_data);
}

/*
 * Create reformat data for encapsulation IPV4 tunnel, and copy it to reformat_data pointer
 *
 * @rule [in]: current rule for encapsulation
 * @sw_sn_inc [in]: if true, sequence number will be incremented in software
 * @reformat_data [out]: pointer to created data
 * @reformat_data_sz [out]: data size
 */
static void create_ipv4_tunnel_encap(struct encrypt_rule *rule,
				     bool sw_sn_inc,
				     uint8_t *reformat_data,
				     uint16_t *reformat_data_sz)
{
	uint8_t reformat_encap_data[50] = {
		0xaa, 0xbb, 0xcc, 0xdd, 0xee, 0xff,			    /* mac_dst */
		0x11, 0x22, 0x33, 0x44, 0x55, 0x66,			    /* mac_src */
		0x08, 0x00,						    /* mac_type */
		0x45, 0x00, 0x00, 0x00, 0x00, 0x00,			    /* IP v4 */
		0x00, 0x00, 0x00, 0x32, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, /* IP src */
		0x00, 0x00, 0x00, 0x00,					    /* IP dst */
		0x00, 0x00, 0x00, 0x00,					    /* SPI */
		0x00, 0x00, 0x00, 0x00,					    /* SN */
		0x00, 0x00, 0x00, 0x00,					    /* IV */
		0x00, 0x00, 0x00, 0x00,
	};

	/* dst IP was already converted to big endian */
	reformat_encap_data[ENCAP_DST_IP_IDX_IP4] = GET_BYTE(rule->encap_dst_ip4, 0);
	reformat_encap_data[ENCAP_DST_IP_IDX_IP4 + 1] = GET_BYTE(rule->encap_dst_ip4, 1);
	reformat_encap_data[ENCAP_DST_IP_IDX_IP4 + 2] = GET_BYTE(rule->encap_dst_ip4, 2);
	reformat_encap_data[ENCAP_DST_IP_IDX_IP4 + 3] = GET_BYTE(rule->encap_dst_ip4, 3);

	reformat_encap_data[ENCAP_ESP_SPI_IDX_TUNNEL_IP4] = GET_BYTE(rule->esp_spi, 3);
	reformat_encap_data[ENCAP_ESP_SPI_IDX_TUNNEL_IP4 + 1] = GET_BYTE(rule->esp_spi, 2);
	reformat_encap_data[ENCAP_ESP_SPI_IDX_TUNNEL_IP4 + 2] = GET_BYTE(rule->esp_spi, 1);
	reformat_encap_data[ENCAP_ESP_SPI_IDX_TUNNEL_IP4 + 3] = GET_BYTE(rule->esp_spi, 0);

	reformat_encap_data[ENCAP_IP_ID_IDX_IP4] = GET_BYTE(current_ip_id, 1);
	reformat_encap_data[ENCAP_IP_ID_IDX_IP4 + 1] = GET_BYTE(current_ip_id, 0);
	++current_ip_id;

	if (sw_sn_inc == true) {
		reformat_encap_data[ENCAP_ESP_SN_IDX_TUNNEL_IP4] = GET_BYTE(rule->current_sn, 3);
		reformat_encap_data[ENCAP_ESP_SN_IDX_TUNNEL_IP4 + 1] = GET_BYTE(rule->current_sn, 2);
		reformat_encap_data[ENCAP_ESP_SN_IDX_TUNNEL_IP4 + 2] = GET_BYTE(rule->current_sn, 1);
		reformat_encap_data[ENCAP_ESP_SN_IDX_TUNNEL_IP4 + 3] = GET_BYTE(rule->current_sn, 0);
	}

	memcpy(reformat_data, reformat_encap_data, sizeof(reformat_encap_data));
	*reformat_data_sz = sizeof(reformat_encap_data);
}

/*
 * Create reformat data for encapsulation IPV6 tunnel, and copy it to reformat_data pointer
 *
 * @rule [in]: current rule for encapsulation
 * @sw_sn_inc [in]: if true, sequence number will be incremented in software
 * @reformat_data [out]: pointer to created data
 * @reformat_data_sz [out]: data size
 */
static void create_ipv6_tunnel_encap(struct encrypt_rule *rule,
				     bool sw_sn_inc,
				     uint8_t *reformat_data,
				     uint16_t *reformat_data_sz)
{
	uint8_t reformat_encap_data[70] = {
		0xaa, 0xbb, 0xcc, 0xdd, 0xee, 0xff,		/* mac_dst */
		0x11, 0x22, 0x33, 0x44, 0x55, 0x66,		/* mac_src */
		0x86, 0xdd,					/* mac_type */
		0x60, 0x00, 0x00, 0x00,				/* IP v6 */
		0x00, 0x00, 0x32, 0x40, 0x02, 0x02, 0x02, 0x02, /* IP src */
		0x02, 0x02, 0x02, 0x02,				/* IP src */
		0x02, 0x02, 0x02, 0x02,				/* IP src */
		0x02, 0x02, 0x02, 0x02,				/* IP src */
		0x01, 0x01, 0x01, 0x01,				/* IP dst */
		0x01, 0x01, 0x01, 0x01,				/* IP dst */
		0x01, 0x01, 0x01, 0x01,				/* IP dst */
		0x01, 0x01, 0x01, 0x01,				/* IP dst */
		0x00, 0x00, 0x00, 0x00,				/* SPI */
		0x00, 0x00, 0x00, 0x00,				/* SN */
		0x00, 0x00, 0x00, 0x00,				/* IV */
		0x00, 0x00, 0x00, 0x00,
	};

	/* dst IP was already converted to big endian */
	reformat_encap_data[ENCAP_DST_IP_IDX_IP6] = GET_BYTE(rule->encap_dst_ip6[0], 0);
	reformat_encap_data[ENCAP_DST_IP_IDX_IP6 + 1] = GET_BYTE(rule->encap_dst_ip6[0], 1);
	reformat_encap_data[ENCAP_DST_IP_IDX_IP6 + 2] = GET_BYTE(rule->encap_dst_ip6[0], 2);
	reformat_encap_data[ENCAP_DST_IP_IDX_IP6 + 3] = GET_BYTE(rule->encap_dst_ip6[0], 3);
	reformat_encap_data[ENCAP_DST_IP_IDX_IP6 + 4] = GET_BYTE(rule->encap_dst_ip6[1], 0);
	reformat_encap_data[ENCAP_DST_IP_IDX_IP6 + 5] = GET_BYTE(rule->encap_dst_ip6[1], 1);
	reformat_encap_data[ENCAP_DST_IP_IDX_IP6 + 6] = GET_BYTE(rule->encap_dst_ip6[1], 2);
	reformat_encap_data[ENCAP_DST_IP_IDX_IP6 + 7] = GET_BYTE(rule->encap_dst_ip6[1], 3);
	reformat_encap_data[ENCAP_DST_IP_IDX_IP6 + 8] = GET_BYTE(rule->encap_dst_ip6[2], 0);
	reformat_encap_data[ENCAP_DST_IP_IDX_IP6 + 9] = GET_BYTE(rule->encap_dst_ip6[2], 1);
	reformat_encap_data[ENCAP_DST_IP_IDX_IP6 + 10] = GET_BYTE(rule->encap_dst_ip6[2], 2);
	reformat_encap_data[ENCAP_DST_IP_IDX_IP6 + 11] = GET_BYTE(rule->encap_dst_ip6[2], 3);
	reformat_encap_data[ENCAP_DST_IP_IDX_IP6 + 12] = GET_BYTE(rule->encap_dst_ip6[3], 0);
	reformat_encap_data[ENCAP_DST_IP_IDX_IP6 + 13] = GET_BYTE(rule->encap_dst_ip6[3], 1);
	reformat_encap_data[ENCAP_DST_IP_IDX_IP6 + 14] = GET_BYTE(rule->encap_dst_ip6[3], 2);
	reformat_encap_data[ENCAP_DST_IP_IDX_IP6 + 15] = GET_BYTE(rule->encap_dst_ip6[3], 3);

	reformat_encap_data[ENCAP_ESP_SPI_IDX_TUNNEL_IP6] = GET_BYTE(rule->esp_spi, 3);
	reformat_encap_data[ENCAP_ESP_SPI_IDX_TUNNEL_IP6 + 1] = GET_BYTE(rule->esp_spi, 2);
	reformat_encap_data[ENCAP_ESP_SPI_IDX_TUNNEL_IP6 + 2] = GET_BYTE(rule->esp_spi, 1);
	reformat_encap_data[ENCAP_ESP_SPI_IDX_TUNNEL_IP6 + 3] = GET_BYTE(rule->esp_spi, 0);

	if (sw_sn_inc == true) {
		reformat_encap_data[ENCAP_ESP_SN_IDX_TUNNEL_IP6] = GET_BYTE(rule->current_sn, 3);
		reformat_encap_data[ENCAP_ESP_SN_IDX_TUNNEL_IP6 + 1] = GET_BYTE(rule->current_sn, 2);
		reformat_encap_data[ENCAP_ESP_SN_IDX_TUNNEL_IP6 + 2] = GET_BYTE(rule->current_sn, 1);
		reformat_encap_data[ENCAP_ESP_SN_IDX_TUNNEL_IP6 + 3] = GET_BYTE(rule->current_sn, 0);
	}

	memcpy(reformat_data, reformat_encap_data, sizeof(reformat_encap_data));
	*reformat_data_sz = sizeof(reformat_encap_data);
}

/*
 * Create egress IP classifier to choose to which encrypt pipe to forward the packet
 *
 * @port [in]: port of the pipe
 * @is_root [in]: true for vnf mode
 * @debug_mode [in]: true for vnf mode
 * @encrypt_pipes [in]: all the encrypt pipes
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t create_egress_ip_classifier(struct doca_flow_port *port,
						bool is_root,
						bool debug_mode,
						struct encrypt_pipes *encrypt_pipes)
{
	struct doca_flow_match match;
	struct doca_flow_monitor monitor;
	struct doca_flow_fwd fwd;
	struct doca_flow_pipe_cfg *pipe_cfg;
	struct security_gateway_pipe_info *pipe = &encrypt_pipes->egress_ip_classifier;
	struct doca_flow_pipe_entry **entry = NULL;
	doca_error_t result;
	struct entries_status status;
	int num_of_entries = 2;

	memset(&match, 0, sizeof(match));
	memset(&monitor, 0, sizeof(monitor));
	memset(&fwd, 0, sizeof(fwd));
	memset(&status, 0, sizeof(status));

	match.parser_meta.outer_l3_type = UINT32_MAX;
	result = doca_flow_pipe_cfg_create(&pipe_cfg, port);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create doca_flow_pipe_cfg: %s", doca_error_get_descr(result));
		return result;
	}

	result = doca_flow_pipe_cfg_set_name(pipe_cfg, "IP_CLASSIFIER_PIPE");
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg name: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_type(pipe_cfg, DOCA_FLOW_PIPE_BASIC);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg type: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_is_root(pipe_cfg, is_root);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg is_root: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_domain(pipe_cfg, DOCA_FLOW_PIPE_DOMAIN_SECURE_EGRESS);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg domain: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_dir_info(pipe_cfg, DOCA_FLOW_DIRECTION_HOST_TO_NETWORK);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg dir_info: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_match(pipe_cfg, &match, NULL);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg match: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	if (debug_mode) {
		monitor.counter_type = DOCA_FLOW_RESOURCE_TYPE_NON_SHARED;
		result = doca_flow_pipe_cfg_set_monitor(pipe_cfg, &monitor);
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg monitor: %s", doca_error_get_descr(result));
			goto destroy_pipe_cfg;
		}
	}

	fwd.type = DOCA_FLOW_FWD_PIPE;

	result = doca_flow_pipe_create(pipe_cfg, &fwd, NULL, &pipe->pipe);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create IP classifier pipe: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}

	doca_flow_pipe_cfg_destroy(pipe_cfg);

	if (debug_mode) {
		pipe->entries_info =
			(struct security_gateway_entry_info *)calloc(2, sizeof(struct security_gateway_entry_info));
		if (pipe->entries_info == NULL) {
			DOCA_LOG_ERR("Failed to allocate entries array");
			return DOCA_ERROR_NO_MEMORY;
		}
	}

	match.parser_meta.outer_l3_type = DOCA_FLOW_L3_META_IPV4;
	fwd.next_pipe = encrypt_pipes->ipv4_encrypt_pipe.pipe;

	if (debug_mode) {
		snprintf(pipe->entries_info[pipe->nb_entries].name, MAX_NAME_LEN, "IPv4");
		entry = &pipe->entries_info[pipe->nb_entries++].entry;
	}
	result = doca_flow_pipe_add_entry(0,
					  pipe->pipe,
					  &match,
					  NULL,
					  NULL,
					  &fwd,
					  DOCA_FLOW_WAIT_FOR_BATCH,
					  &status,
					  entry);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to add ipv4 entry: %s", doca_error_get_descr(result));
		return result;
	}

	match.parser_meta.outer_l3_type = DOCA_FLOW_L3_META_IPV6;
	fwd.next_pipe = encrypt_pipes->ipv6_encrypt_pipe.pipe;

	if (debug_mode) {
		snprintf(pipe->entries_info[pipe->nb_entries].name, MAX_NAME_LEN, "IPv6");
		entry = &pipe->entries_info[pipe->nb_entries++].entry;
	}
	result = doca_flow_pipe_add_entry(0, pipe->pipe, &match, NULL, NULL, &fwd, DOCA_FLOW_NO_WAIT, &status, entry);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to add ipv6 entry: %s", doca_error_get_descr(result));
		return result;
	}

	result = doca_flow_entries_process(port, 0, DEFAULT_TIMEOUT_US, num_of_entries);
	if (result != DOCA_SUCCESS)
		return result;
	if (status.nb_processed != num_of_entries || status.failure)
		return DOCA_ERROR_BAD_STATE;

	return DOCA_SUCCESS;

destroy_pipe_cfg:
	doca_flow_pipe_cfg_destroy(pipe_cfg);
	return result;
}

/*
 * Create ipsec encrypt pipe changeable meta data match and changeable shared IPSEC encryption object
 *
 * @port [in]: port of the pipe
 * @port_id [in]: port ID to forward the packet to
 * @expected_entries [in]: expected number of entries
 * @app_cfg [in]: application configuration struct
 * @l3_type [in]: DOCA_FLOW_L3_TYPE_IP4 / DOCA_FLOW_L3_TYPE_IP6
 * @pipe_info [out]: pipe info struct
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t create_ipsec_encrypt_pipe(struct doca_flow_port *port,
					      uint16_t port_id,
					      int expected_entries,
					      struct ipsec_security_gw_config *app_cfg,
					      enum doca_flow_l3_type l3_type,
					      struct security_gateway_pipe_info *pipe_info)
{
	int nb_actions = 2;
	struct doca_flow_match match;
	struct doca_flow_match match_mask;
	struct doca_flow_monitor monitor;
	struct doca_flow_actions actions, actions_arr[nb_actions];
	struct doca_flow_fwd fwd;
	struct doca_flow_pipe_cfg *pipe_cfg;
	doca_error_t result;
	union security_gateway_pkt_meta meta = {0};

	memset(&match, 0, sizeof(match));
	memset(&match_mask, 0, sizeof(match_mask));
	memset(&monitor, 0, sizeof(monitor));
	memset(&actions, 0, sizeof(actions));
	memset(&fwd, 0, sizeof(fwd));

	meta.rule_id = -1;
	match_mask.meta.pkt_meta = meta.u32;
	match.meta.pkt_meta = 0xffffffff;
	match_mask.outer.l3_type = l3_type;
	match.outer.l3_type = l3_type;

	if (app_cfg->offload == IPSEC_SECURITY_GW_ESP_OFFLOAD_BOTH ||
	    app_cfg->offload == IPSEC_SECURITY_GW_ESP_OFFLOAD_ENCAP)
		actions.has_crypto_encap = true;

	actions.crypto_encap.action_type = DOCA_FLOW_CRYPTO_REFORMAT_ENCAP;
	actions.crypto_encap.icv_size = 0xffff;
	actions.crypto.resource_type = DOCA_FLOW_CRYPTO_RESOURCE_IPSEC_SA;
	if (!app_cfg->sw_sn_inc_enable) {
		actions.crypto.ipsec_sa.sn_en = !app_cfg->sw_sn_inc_enable;
	}
	actions.crypto.action_type = DOCA_FLOW_CRYPTO_ACTION_ENCRYPT;
	actions.crypto.crypto_id = ENCRYPT_DUMMY_ID;

	if (app_cfg->mode == IPSEC_SECURITY_GW_TUNNEL) {
		actions.crypto_encap.net_type = DOCA_FLOW_CRYPTO_HEADER_ESP_TUNNEL;
		actions_arr[0] = actions;
		actions_arr[1] = actions;
		/* action idx 0 will add encap ipv4 */
		memset(actions_arr[0].crypto_encap.encap_data, 0xff, 50);
		actions_arr[0].crypto_encap.data_size = 50;
		/* action idx 1 will add encap ipv6 */
		memset(actions_arr[1].crypto_encap.encap_data, 0xff, 70);
		actions_arr[1].crypto_encap.data_size = 70;
	} else if (app_cfg->mode == IPSEC_SECURITY_GW_TRANSPORT) {
		actions.crypto_encap.net_type = (l3_type == DOCA_FLOW_L3_TYPE_IP4) ?
							DOCA_FLOW_CRYPTO_HEADER_ESP_OVER_IPV4 :
							DOCA_FLOW_CRYPTO_HEADER_ESP_OVER_IPV6;
		memset(actions.crypto_encap.encap_data, 0xff, 16);
		actions.crypto_encap.data_size = 16;
		actions_arr[0] = actions;
	} else {
		actions.crypto_encap.net_type = (l3_type == DOCA_FLOW_L3_TYPE_IP4) ?
							DOCA_FLOW_CRYPTO_HEADER_UDP_ESP_OVER_IPV4 :
							DOCA_FLOW_CRYPTO_HEADER_UDP_ESP_OVER_IPV6;
		memset(actions.crypto_encap.encap_data, 0xff, 24);
		actions.crypto_encap.data_size = 24;
		actions_arr[0] = actions;
	}

	struct doca_flow_actions *actions_list[] = {&actions_arr[0], &actions_arr[1]};

	result = doca_flow_pipe_cfg_create(&pipe_cfg, port);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create doca_flow_pipe_cfg: %s", doca_error_get_descr(result));
		return result;
	}

	result = doca_flow_pipe_cfg_set_name(pipe_cfg, "ENCRYPT_PIPE");
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg name: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_type(pipe_cfg, DOCA_FLOW_PIPE_BASIC);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg type: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_enable_strict_matching(pipe_cfg, true);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg enable_strict_matching: %s",
			     doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_domain(pipe_cfg, DOCA_FLOW_PIPE_DOMAIN_SECURE_EGRESS);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg domain: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_dir_info(pipe_cfg, DOCA_FLOW_DIRECTION_HOST_TO_NETWORK);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg dir_info: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_nr_entries(pipe_cfg, expected_entries);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg nr_entries: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_match(pipe_cfg, &match, &match_mask);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg match: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	/* 2 actions only in tunnel mode */
	result =
		doca_flow_pipe_cfg_set_actions(pipe_cfg,
					       actions_list,
					       NULL,
					       NULL,
					       app_cfg->mode == IPSEC_SECURITY_GW_TUNNEL ? nb_actions : nb_actions - 1);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg actions: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	if (app_cfg->debug_mode) {
		monitor.counter_type = DOCA_FLOW_RESOURCE_TYPE_NON_SHARED;
		result = doca_flow_pipe_cfg_set_monitor(pipe_cfg, &monitor);
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg monitor: %s", doca_error_get_descr(result));
			goto destroy_pipe_cfg;
		}
	}

	fwd.type = DOCA_FLOW_FWD_PORT;
	fwd.port_id = port_id;

	result = doca_flow_pipe_create(pipe_cfg, &fwd, NULL, &pipe_info->pipe);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create encrypt pipe: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}

	if (app_cfg->debug_mode) {
		pipe_info->entries_info =
			(struct security_gateway_entry_info *)calloc(expected_entries,
								     sizeof(struct security_gateway_entry_info));
		if (pipe_info->entries_info == NULL) {
			DOCA_LOG_ERR("Failed to allocate entries array");
			result = DOCA_ERROR_NO_MEMORY;
		}
	}

destroy_pipe_cfg:
	doca_flow_pipe_cfg_destroy(pipe_cfg);
	return result;
}

/*
 * Create pipe with source ipv6 match, and fwd to the hairpin pipe
 *
 * @port [in]: port of the pipe
 * @debug_mode [in]: true if running in debug mode
 * @expected_entries [in]: expected number of entries
 * @protocol_type [in]: DOCA_FLOW_L4_TYPE_EXT_TCP / DOCA_FLOW_L4_TYPE_EXT_UDP
 * @hairpin_pipe [in]: pipe to forward the packets
 * @pipe_info [out]: pipe info struct
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t create_ipsec_src_ip6_pipe(struct doca_flow_port *port,
					      bool debug_mode,
					      int expected_entries,
					      enum doca_flow_l4_type_ext protocol_type,
					      struct doca_flow_pipe *hairpin_pipe,
					      struct security_gateway_pipe_info *pipe_info)
{
	int nb_actions = 1;
	struct doca_flow_match match;
	struct doca_flow_monitor monitor;
	struct doca_flow_fwd fwd;
	struct doca_flow_actions actions, *actions_arr[nb_actions];
	struct doca_flow_pipe_cfg *pipe_cfg;
	doca_error_t result;

	memset(&match, 0, sizeof(match));
	memset(&monitor, 0, sizeof(monitor));
	memset(&fwd, 0, sizeof(fwd));
	memset(&actions, 0, sizeof(actions));

	match.parser_meta.outer_l3_type = DOCA_FLOW_L3_META_IPV6;
	match.outer.l3_type = DOCA_FLOW_L3_TYPE_IP6;
	match.parser_meta.outer_l4_type = (protocol_type == DOCA_FLOW_L4_TYPE_EXT_TCP) ? DOCA_FLOW_L4_META_TCP :
											 DOCA_FLOW_L4_META_UDP;
	SET_IP6_ADDR(match.outer.ip6.src_ip, 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff);

	actions.meta.u32[0] = UINT32_MAX;
	actions_arr[0] = &actions;

	result = doca_flow_pipe_cfg_create(&pipe_cfg, port);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create doca_flow_pipe_cfg: %s", doca_error_get_descr(result));
		return result;
	}

	result = doca_flow_pipe_cfg_set_name(pipe_cfg, "SRC_IP6_PIPE");
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg name: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_type(pipe_cfg, DOCA_FLOW_PIPE_BASIC);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg type: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_is_root(pipe_cfg, false);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg is_root: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_nr_entries(pipe_cfg, expected_entries);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg nr_entries: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_match(pipe_cfg, &match, NULL);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg match: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_actions(pipe_cfg, actions_arr, NULL, NULL, nb_actions);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg actions: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	if (debug_mode) {
		monitor.counter_type = DOCA_FLOW_RESOURCE_TYPE_NON_SHARED;
		result = doca_flow_pipe_cfg_set_monitor(pipe_cfg, &monitor);
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg monitor: %s", doca_error_get_descr(result));
			goto destroy_pipe_cfg;
		}
	}

	fwd.type = DOCA_FLOW_FWD_PIPE;
	fwd.next_pipe = hairpin_pipe;

	result = doca_flow_pipe_create(pipe_cfg, &fwd, NULL, &pipe_info->pipe);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create hairpin pipe: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}

	if (debug_mode) {
		pipe_info->entries_info =
			(struct security_gateway_entry_info *)calloc(expected_entries,
								     sizeof(struct security_gateway_entry_info));
		if (pipe_info->entries_info == NULL) {
			DOCA_LOG_ERR("Failed to allocate entries array");
			result = DOCA_ERROR_NO_MEMORY;
		}
	}

destroy_pipe_cfg:
	doca_flow_pipe_cfg_destroy(pipe_cfg);
	return result;
}

/*
 * Create pipe with 5 tuple match, changeable set meta, and fwd to the second port
 *
 * @port [in]: port of the pipe
 * @debug_mode [in]: true if running in debug mode
 * @expected_entries [in]: expected number of entries
 * @protocol_type [in]: DOCA_FLOW_L4_TYPE_EXT_TCP / DOCA_FLOW_L4_TYPE_EXT_UDP
 * @l3_type [in]: DOCA_FLOW_L3_TYPE_IP4 / DOCA_FLOW_L3_TYPE_IP6
 * @fwd [in]: pointer to forward struct
 * @pipe_info [out]: pipe info struct
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t create_ipsec_hairpin_pipe(struct doca_flow_port *port,
					      bool debug_mode,
					      int expected_entries,
					      enum doca_flow_l4_type_ext protocol_type,
					      enum doca_flow_l3_type l3_type,
					      struct doca_flow_fwd *fwd,
					      struct security_gateway_pipe_info *pipe_info)
{
	int nb_actions = 1;
	struct doca_flow_match match;
	struct doca_flow_monitor monitor;
	struct doca_flow_actions actions, *actions_arr[nb_actions];
	struct doca_flow_pipe_cfg *pipe_cfg;
	doca_error_t result;

	memset(&match, 0, sizeof(match));
	memset(&monitor, 0, sizeof(monitor));
	memset(&actions, 0, sizeof(actions));

	match.outer.l4_type_ext = protocol_type;
	match.outer.l3_type = l3_type;
	if (l3_type == DOCA_FLOW_L3_TYPE_IP4) {
		match.outer.ip4.dst_ip = 0xffffffff;
		match.outer.ip4.src_ip = 0xffffffff;
	} else {
		match.meta.u32[0] = UINT32_MAX;
		SET_IP6_ADDR(match.outer.ip6.dst_ip, 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff);
	}

	SET_L4_PORT(outer, src_port, 0xffff);
	SET_L4_PORT(outer, dst_port, 0xffff);

	actions.meta.pkt_meta = 0xffffffff;
	actions_arr[0] = &actions;

	result = doca_flow_pipe_cfg_create(&pipe_cfg, port);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create doca_flow_pipe_cfg: %s", doca_error_get_descr(result));
		return result;
	}

	result = doca_flow_pipe_cfg_set_name(pipe_cfg, "HAIRPIN_PIPE");
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg name: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_type(pipe_cfg, DOCA_FLOW_PIPE_BASIC);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg type: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_is_root(pipe_cfg, false);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg is_root: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_nr_entries(pipe_cfg, expected_entries);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg nr_entries: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_match(pipe_cfg, &match, &match);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg match: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_actions(pipe_cfg, actions_arr, NULL, NULL, nb_actions);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg actions: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	if (debug_mode) {
		monitor.counter_type = DOCA_FLOW_RESOURCE_TYPE_NON_SHARED;
		result = doca_flow_pipe_cfg_set_monitor(pipe_cfg, &monitor);
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg monitor: %s", doca_error_get_descr(result));
			goto destroy_pipe_cfg;
		}
	}

	result = doca_flow_pipe_create(pipe_cfg, fwd, NULL, &pipe_info->pipe);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create hairpin pipe: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}

	if (debug_mode) {
		pipe_info->entries_info =
			(struct security_gateway_entry_info *)calloc(expected_entries,
								     sizeof(struct security_gateway_entry_info));
		if (pipe_info->entries_info == NULL) {
			DOCA_LOG_ERR("Failed to allocate entries array");
			result = DOCA_ERROR_NO_MEMORY;
		}
	}

destroy_pipe_cfg:
	doca_flow_pipe_cfg_destroy(pipe_cfg);
	return result;
}

/*
 * Create control pipe for unsecured port
 *
 * @port [in]: port of the pipe
 * @is_root [in]: true in vnf mode
 * @debug_mode [in]: true if running in debug mode
 * @pipe_info [out]: pipe info struct
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t create_control_pipe(struct doca_flow_port *port,
					bool is_root,
					bool debug_mode,
					struct security_gateway_pipe_info *pipe_info)
{
	struct doca_flow_pipe_cfg *pipe_cfg;
	doca_error_t result;

	result = doca_flow_pipe_cfg_create(&pipe_cfg, port);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create doca_flow_pipe_cfg: %s", doca_error_get_descr(result));
		return result;
	}

	result = doca_flow_pipe_cfg_set_name(pipe_cfg, "CONTROL_PIPE");
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg name: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_type(pipe_cfg, DOCA_FLOW_PIPE_CONTROL);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg type: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_is_root(pipe_cfg, is_root);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg is_root: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}
	result = doca_flow_pipe_cfg_set_enable_strict_matching(pipe_cfg, true);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg enable_strict_matching: %s",
			     doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}

	result = doca_flow_pipe_create(pipe_cfg, NULL, NULL, &pipe_info->pipe);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create control pipe: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}

	if (debug_mode) {
		pipe_info->entries_info =
			(struct security_gateway_entry_info *)calloc(4, sizeof(struct security_gateway_entry_info));
		if (pipe_info->entries_info == NULL) {
			DOCA_LOG_ERR("Failed to allocate entries array");
			result = DOCA_ERROR_NO_MEMORY;
		}
	}

destroy_pipe_cfg:
	doca_flow_pipe_cfg_destroy(pipe_cfg);
	return result;
}

/*
 * Check if BW perf is enabled
 *
 * @app_cfg [in]: application configuration struct
 * @return: true if BW perf is enabled.
 */
static inline bool is_perf_bw(struct ipsec_security_gw_config *app_cfg)
{
	return (app_cfg->perf_measurement == IPSEC_SECURITY_GW_PERF_BW ||
		app_cfg->perf_measurement == IPSEC_SECURITY_GW_PERF_BOTH);
}

/*
 * Add control pipe entries
 * - entry that forwards IPv4 TCP traffic to IPv4 TCP pipe,
 * - entry that forwards IPv4 UDP traffic to IPv4 UDP pipe,
 * - entry that forwards IPv6 TCP traffic to source IPv6 TCP pipe,
 * - entry that forwards IPv6 UDP traffic to source IPv6 UDP pipe,
 * - entry with lower priority that drop the packets
 *
 * @control_pipe [in]: control pipe pointer
 * @pipes [in]: all the pipes to forward the packets to
 * @perf_bw [in]: true if perf mode includes bandwidth
 * @debug_mode [in]: true if running in debug mode
 * @is_root [in]: true in vnf mode
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t add_control_pipe_entries(struct security_gateway_pipe_info *control_pipe,
					     struct encrypt_pipes *pipes,
					     bool perf_bw,
					     bool debug_mode,
					     bool is_root)
{
	struct doca_flow_pipe_entry **entry = NULL;
	struct doca_flow_monitor monitor;
	struct doca_flow_monitor *monitor_ptr = NULL;
	struct doca_flow_match match;
	struct doca_flow_fwd fwd;

	doca_error_t result;

	memset(&match, 0, sizeof(match));
	memset(&monitor, 0, sizeof(monitor));
	memset(&fwd, 0, sizeof(fwd));

	if (debug_mode && !is_root) {
		monitor.counter_type = DOCA_FLOW_RESOURCE_TYPE_NON_SHARED;
		monitor_ptr = &monitor;
	}
	match.outer.l3_type = DOCA_FLOW_L3_TYPE_IP4;
	match.outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_TCP;

	fwd.type = DOCA_FLOW_FWD_PIPE;
	fwd.next_pipe = pipes->ipv4_tcp_pipe.pipe;
	if (debug_mode) {
		snprintf(control_pipe->entries_info[control_pipe->nb_entries].name, MAX_NAME_LEN, "ipv4_tcp");
		entry = &control_pipe->entries_info[control_pipe->nb_entries++].entry;
	}
	result = doca_flow_pipe_control_add_entry(0,
						  0,
						  control_pipe->pipe,
						  &match,
						  NULL,
						  NULL,
						  NULL,
						  NULL,
						  NULL,
						  monitor_ptr,
						  &fwd,
						  NULL,
						  entry);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to add TCP IPv4 control entry: %s", doca_error_get_descr(result));
		return result;
	}

	match.outer.l3_type = DOCA_FLOW_L3_TYPE_IP4;
	match.outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_UDP;

	fwd.type = DOCA_FLOW_FWD_PIPE;
	fwd.next_pipe = pipes->ipv4_udp_pipe.pipe;
	if (debug_mode) {
		snprintf(control_pipe->entries_info[control_pipe->nb_entries].name, MAX_NAME_LEN, "ipv4_udp");
		entry = &control_pipe->entries_info[control_pipe->nb_entries++].entry;
	}
	result = doca_flow_pipe_control_add_entry(0,
						  0,
						  control_pipe->pipe,
						  &match,
						  NULL,
						  NULL,
						  NULL,
						  NULL,
						  NULL,
						  monitor_ptr,
						  &fwd,
						  NULL,
						  entry);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to add UDP IPv4 control entry: %s", doca_error_get_descr(result));
		return result;
	}

	match.outer.l3_type = DOCA_FLOW_L3_TYPE_IP6;
	match.outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_TCP;

	fwd.type = DOCA_FLOW_FWD_PIPE;
	if (!perf_bw)
		fwd.next_pipe = pipes->ipv6_src_tcp_pipe.pipe;
	else
		fwd.next_pipe = pipes->ipv6_tcp_pipe.pipe;
	if (debug_mode) {
		snprintf(control_pipe->entries_info[control_pipe->nb_entries].name, MAX_NAME_LEN, "ipv6_tcp");
		entry = &control_pipe->entries_info[control_pipe->nb_entries++].entry;
	}
	result = doca_flow_pipe_control_add_entry(0,
						  0,
						  control_pipe->pipe,
						  &match,
						  NULL,
						  NULL,
						  NULL,
						  NULL,
						  NULL,
						  monitor_ptr,
						  &fwd,
						  NULL,
						  entry);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to add TCP IPv6 control entry: %s", doca_error_get_descr(result));
		return result;
	}

	match.outer.l3_type = DOCA_FLOW_L3_TYPE_IP6;
	match.outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_UDP;

	fwd.type = DOCA_FLOW_FWD_PIPE;
	if (!perf_bw)
		fwd.next_pipe = pipes->ipv6_src_udp_pipe.pipe;
	else
		fwd.next_pipe = pipes->ipv6_udp_pipe.pipe;
	if (debug_mode) {
		snprintf(control_pipe->entries_info[control_pipe->nb_entries].name, MAX_NAME_LEN, "ipv6_udp");
		entry = &control_pipe->entries_info[control_pipe->nb_entries++].entry;
	}
	result = doca_flow_pipe_control_add_entry(0,
						  0,
						  control_pipe->pipe,
						  &match,
						  NULL,
						  NULL,
						  NULL,
						  NULL,
						  NULL,
						  monitor_ptr,
						  &fwd,
						  NULL,
						  entry);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to add UDP IPv6 control entry: %s", doca_error_get_descr(result));
		return result;
	}

	return DOCA_SUCCESS;
}

/*
 * Update the crypto config for encrypt transport mode according to the rule
 *
 * @crypto_cfg [in]: shared object config
 * @rule [in]: encrypt rule
 */
static void create_ipsec_encrypt_shared_object_transport(struct doca_flow_crypto_encap_action *crypto_cfg,
							 struct encrypt_rule *rule)
{
	create_transport_encap(rule, false, crypto_cfg->encap_data, &crypto_cfg->data_size);
}

/*
 * Update the crypto config for encrypt transport over UDP mode according to the rule
 *
 * @crypto_cfg [in]: shared object config
 * @rule [in]: encrypt rule
 */
static void create_ipsec_encrypt_shared_object_transport_over_udp(struct doca_flow_crypto_encap_action *crypto_cfg,
								  struct encrypt_rule *rule)
{
	create_udp_transport_encap(rule, false, crypto_cfg->encap_data, &crypto_cfg->data_size);
}

/*
 * Update the crypto config for encrypt trunnel mode according to the rule
 *
 * @crypto_cfg [in]: shared object config
 * @rule [in]: encrypt rule
 */
static void create_ipsec_encrypt_shared_object_tunnel(struct doca_flow_crypto_encap_action *crypto_cfg,
						      struct encrypt_rule *rule)
{
	if (rule->encap_l3_type == DOCA_FLOW_L3_TYPE_IP4)
		create_ipv4_tunnel_encap(rule, false, crypto_cfg->encap_data, &crypto_cfg->data_size);
	else
		create_ipv6_tunnel_encap(rule, false, crypto_cfg->encap_data, &crypto_cfg->data_size);
}

/*
 * Config and bind shared IPSEC object for encryption
 *
 * @app_sa_attrs [in]: SA attributes
 * @app_cfg [in]: application configuration struct
 * @ipsec_id [in]: shared object ID
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t create_ipsec_encrypt_shared_object(struct ipsec_security_gw_sa_attrs *app_sa_attrs,
						       struct ipsec_security_gw_config *app_cfg,
						       uint32_t ipsec_id)
{
	struct doca_flow_shared_resource_cfg cfg;
	doca_error_t result;

	memset(&cfg, 0, sizeof(cfg));

	cfg.domain = DOCA_FLOW_PIPE_DOMAIN_SECURE_EGRESS;
	cfg.ipsec_sa_cfg.icv_len = app_sa_attrs->icv_length;
	cfg.ipsec_sa_cfg.salt = app_sa_attrs->salt;
	cfg.ipsec_sa_cfg.key_cfg.key_type = app_sa_attrs->key_type;
	cfg.ipsec_sa_cfg.key_cfg.key = (void *)&app_sa_attrs->enc_key_data;
	if (!app_cfg->sw_sn_inc_enable) {
		cfg.ipsec_sa_cfg.sn_offload_type = DOCA_FLOW_CRYPTO_SN_OFFLOAD_INC;
		cfg.ipsec_sa_cfg.sn_initial = app_cfg->sn_initial;
	}
	/* config ipsec object */
	result = doca_flow_shared_resource_cfg(DOCA_FLOW_SHARED_RESOURCE_IPSEC_SA, ipsec_id, &cfg);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to cfg shared ipsec object: %s", doca_error_get_descr(result));
		return result;
	}

	return DOCA_SUCCESS;
}

/*
 * Create dummy SA, and config and bind doca flow shared resource with it
 *
 * @cfg [in]: application configuration struct
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t create_ipsec_encrypt_dummy_shared_object(struct ipsec_security_gw_config *cfg)
{
	doca_error_t result;
	struct ipsec_security_gw_sa_attrs dummy_sa_attr = {
		.icv_length = DOCA_FLOW_CRYPTO_ICV_LENGTH_16,
		.key_type = DOCA_FLOW_CRYPTO_KEY_256,
		.enc_key_data[0] = 0x01,
		.salt = 0x12345678,
	};

	result = create_ipsec_encrypt_shared_object(&dummy_sa_attr, cfg, ENCRYPT_DUMMY_ID);
	if (result != DOCA_SUCCESS)
		return result;

	return DOCA_SUCCESS;
}

/*
 * Get the relevant pipe for adding the rule
 *
 * @rule [in]: the rule that need to add
 * @pipes [in]: encrypt pipes struct
 * @src_ip6 [in]: true if we want to get the source ipv6 pipe
 * @pipe [out]: output pipe
 */
static void get_pipe_for_rule(struct encrypt_rule *rule,
			      struct encrypt_pipes *pipes,
			      bool src_ip6,
			      struct security_gateway_pipe_info **pipe)
{
	if (!src_ip6) {
		if (rule->l3_type == DOCA_FLOW_L3_TYPE_IP4) {
			if (rule->protocol == DOCA_FLOW_L4_TYPE_EXT_TCP)
				*pipe = &pipes->ipv4_tcp_pipe;
			else
				*pipe = &pipes->ipv4_udp_pipe;
		} else {
			if (rule->protocol == DOCA_FLOW_L4_TYPE_EXT_TCP)
				*pipe = &pipes->ipv6_tcp_pipe;
			else
				*pipe = &pipes->ipv6_udp_pipe;
		}
	} else {
		if (rule->protocol == DOCA_FLOW_L4_TYPE_EXT_TCP)
			*pipe = &pipes->ipv6_src_tcp_pipe;
		else
			*pipe = &pipes->ipv6_src_udp_pipe;
	}
}

/*
 * Add entry to source IPv6 pipe
 *
 * @port [in]: port of the pipe
 * @rule [in]: encrypt rule
 * @pipes [in]: encrypt pipes struct
 * @hairpin_status [in]: the entries status
 * @src_ip_id [in]: source IP unique ID
 * @queue_id [in]: queue ID to forward the packets to
 * @debug_mode [in]: true when running in debug mode
 * @i [in]: rule id
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t add_src_ip6_entry(struct doca_flow_port *port,
				      struct encrypt_rule *rule,
				      struct encrypt_pipes *pipes,
				      struct entries_status *hairpin_status,
				      uint32_t src_ip_id,
				      uint16_t queue_id,
				      bool debug_mode,
				      int i)
{
	struct doca_flow_match match;
	struct doca_flow_actions actions;
	enum doca_flow_flags_type flags;
	struct security_gateway_pipe_info *pipe;
	struct doca_flow_pipe_entry **entry = NULL;
	doca_error_t result;

	memset(&match, 0, sizeof(match));
	memset(&actions, 0, sizeof(actions));

	get_pipe_for_rule(rule, pipes, true, &pipe);

	memcpy(match.outer.ip6.src_ip, rule->ip6.src_ip, sizeof(rule->ip6.src_ip));
	actions.meta.u32[0] = src_ip_id;

	if (hairpin_status->entries_in_queue == QUEUE_DEPTH - 1)
		flags = DOCA_FLOW_NO_WAIT;
	else
		flags = DOCA_FLOW_WAIT_FOR_BATCH;

	/* add entry to hairpin pipe*/
	if (debug_mode) {
		snprintf(pipe->entries_info[pipe->nb_entries].name, MAX_NAME_LEN, "rule%d", i);
		entry = &pipe->entries_info[pipe->nb_entries++].entry;
	}
	result = doca_flow_pipe_add_entry(queue_id,
					  pipe->pipe,
					  &match,
					  &actions,
					  NULL,
					  NULL,
					  flags,
					  hairpin_status,
					  entry);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to add hairpin pipe entry: %s", doca_error_get_descr(result));
		return result;
	}
	hairpin_status->entries_in_queue++;
	if (hairpin_status->entries_in_queue == QUEUE_DEPTH) {
		result = process_entries(port, hairpin_status, DEFAULT_TIMEOUT_US, queue_id);
		if (result != DOCA_SUCCESS)
			return result;
	}
	return DOCA_SUCCESS;
}

/*
 * Add 5-tuple entries based on a rule
 * If ipv6 - add the source IP to different pipe
 *
 * @port [in]: port of the pipe
 * @rule [in]: encrypt rule
 * @app_cfg [in]: application configuration structure
 * @nb_rules [in]: number of encryption rules
 * @i [in]: rule index
 * @queue_id [in]: queue ID to forward the packets to
 * @hairpin_status [in]: the entries status
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t add_five_tuple_match_entry(struct doca_flow_port *port,
					       struct encrypt_rule *rule,
					       struct ipsec_security_gw_config *app_cfg,
					       int nb_rules,
					       int i,
					       uint16_t queue_id,
					       struct entries_status *hairpin_status)
{
	struct doca_flow_match match;
	struct doca_flow_actions actions;
	struct security_gateway_pipe_info *pipe;
	enum doca_flow_flags_type flags;
	int src_ip_id = 0;
	doca_error_t result;
	union security_gateway_pkt_meta meta = {0};
	struct doca_flow_pipe_entry **entry = NULL;

	memset(&match, 0, sizeof(match));
	memset(&actions, 0, sizeof(actions));

	if (rule->l3_type == DOCA_FLOW_L3_TYPE_IP6 && !is_perf_bw(app_cfg)) {
		src_ip_id = rte_hash_lookup(app_cfg->ip6_table, (void *)&rule->ip6.dst_ip);
		if (src_ip_id < 0) {
			DOCA_LOG_ERR("Failed to find source IP in table");
			return DOCA_ERROR_NOT_FOUND;
		}
		result = add_src_ip6_entry(port,
					   rule,
					   &app_cfg->encrypt_pipes,
					   hairpin_status,
					   src_ip_id,
					   queue_id,
					   app_cfg->debug_mode,
					   i);
		if (result != DOCA_SUCCESS)
			return result;
	}

	get_pipe_for_rule(rule, &app_cfg->encrypt_pipes, false, &pipe);

	match.outer.l4_type_ext = rule->protocol;
	SET_L4_PORT(outer, src_port, rte_cpu_to_be_16(rule->src_port));
	SET_L4_PORT(outer, dst_port, rte_cpu_to_be_16(rule->dst_port));

	if (rule->l3_type == DOCA_FLOW_L3_TYPE_IP4) {
		match.outer.ip4.dst_ip = rule->ip4.dst_ip;
		match.outer.ip4.src_ip = rule->ip4.src_ip;
	} else {
		match.meta.u32[0] = src_ip_id;
		memcpy(match.outer.ip6.dst_ip, rule->ip6.dst_ip, sizeof(rule->ip6.dst_ip));
	}

	meta.encrypt = 1;
	meta.rule_id = i;
	actions.meta.pkt_meta = meta.u32;
	actions.action_idx = 0;

	if (i == nb_rules - 1 || hairpin_status->entries_in_queue == QUEUE_DEPTH - 1)
		flags = DOCA_FLOW_NO_WAIT;
	else
		flags = DOCA_FLOW_WAIT_FOR_BATCH;

	/* add entry to hairpin pipe*/
	if (app_cfg->debug_mode) {
		snprintf(pipe->entries_info[pipe->nb_entries].name, MAX_NAME_LEN, "rule%d", i);
		entry = &pipe->entries_info[pipe->nb_entries++].entry;
	}
	result = doca_flow_pipe_add_entry(queue_id,
					  pipe->pipe,
					  &match,
					  &actions,
					  NULL,
					  NULL,
					  flags,
					  hairpin_status,
					  entry);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to add hairpin pipe entry: %s", doca_error_get_descr(result));
		return result;
	}
	hairpin_status->entries_in_queue++;
	if (hairpin_status->entries_in_queue == QUEUE_DEPTH) {
		result = process_entries(port, hairpin_status, DEFAULT_TIMEOUT_US, queue_id);
		if (result != DOCA_SUCCESS)
			return result;
	}

	return DOCA_SUCCESS;
}

doca_error_t add_encrypt_entry(struct encrypt_rule *rule,
			       int rule_id,
			       struct ipsec_security_gw_ports_map **ports,
			       struct ipsec_security_gw_config *app_cfg)
{
	struct doca_flow_match match;
	struct doca_flow_actions actions;
	struct doca_flow_pipe_entry **entry = NULL;
	struct security_gateway_pipe_info *encrypt_pipe;
	struct entries_status hairpin_status;
	struct entries_status encrypt_status;
	struct doca_flow_port *secured_port = NULL;
	struct doca_flow_port *unsecured_port = NULL;
	union security_gateway_pkt_meta meta = {0};
	doca_error_t result;

	if (app_cfg->flow_mode == IPSEC_SECURITY_GW_SWITCH) {
		secured_port = doca_flow_port_switch_get(NULL);
		unsecured_port = doca_flow_port_switch_get(NULL);
	} else {
		secured_port = ports[SECURED_IDX]->port;
		unsecured_port = ports[UNSECURED_IDX]->port;
	}

	memset(&hairpin_status, 0, sizeof(hairpin_status));
	memset(&encrypt_status, 0, sizeof(encrypt_status));
	memset(&match, 0, sizeof(match));
	memset(&actions, 0, sizeof(actions));

	if (app_cfg->mode == IPSEC_SECURITY_GW_TUNNEL && (app_cfg->offload == IPSEC_SECURITY_GW_ESP_OFFLOAD_NONE ||
							  app_cfg->offload == IPSEC_SECURITY_GW_ESP_OFFLOAD_DECAP)) {
		/* SW encap in tunnel mode */
		if (rule->encap_l3_type == DOCA_FLOW_L3_TYPE_IP4)
			encrypt_pipe = &app_cfg->encrypt_pipes.ipv4_encrypt_pipe;
		else
			encrypt_pipe = &app_cfg->encrypt_pipes.ipv6_encrypt_pipe;
	} else {
		if (rule->l3_type == DOCA_FLOW_L3_TYPE_IP4)
			encrypt_pipe = &app_cfg->encrypt_pipes.ipv4_encrypt_pipe;
		else
			encrypt_pipe = &app_cfg->encrypt_pipes.ipv6_encrypt_pipe;
	}
	/* add entry to hairpin pipe*/
	result = add_five_tuple_match_entry(unsecured_port, rule, app_cfg, rule_id + 1, rule_id, 0, &hairpin_status);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to add pipe entry: %s", doca_error_get_descr(result));
		return result;
	}

	/* create ipsec shared object */
	result = create_ipsec_encrypt_shared_object(&rule->sa_attrs, app_cfg, rule_id);
	if (result != DOCA_SUCCESS)
		return result;

	memset(&match, 0, sizeof(match));

	meta.rule_id = rule_id;
	match.meta.pkt_meta = meta.u32;

	actions.action_idx = 0;
	actions.crypto.crypto_id = rule_id;

	if (app_cfg->mode == IPSEC_SECURITY_GW_TUNNEL) {
		create_ipsec_encrypt_shared_object_tunnel(&actions.crypto_encap, rule);
		if (rule->encap_l3_type == DOCA_FLOW_L3_TYPE_IP4)
			actions.action_idx = 0;
		else
			actions.action_idx = 1;
	} else if (app_cfg->mode == IPSEC_SECURITY_GW_TRANSPORT)
		create_ipsec_encrypt_shared_object_transport(&actions.crypto_encap, rule);
	else
		create_ipsec_encrypt_shared_object_transport_over_udp(&actions.crypto_encap, rule);
	actions.crypto_encap.icv_size = get_icv_len_int(rule->sa_attrs.icv_length);
	/* add entry to encrypt pipe*/
	if (app_cfg->debug_mode) {
		snprintf(encrypt_pipe->entries_info[encrypt_pipe->nb_entries].name, MAX_NAME_LEN, "rule%d", rule_id);
		entry = &encrypt_pipe->entries_info[encrypt_pipe->nb_entries++].entry;
	}
	result = doca_flow_pipe_add_entry(0,
					  encrypt_pipe->pipe,
					  &match,
					  &actions,
					  NULL,
					  NULL,
					  DOCA_FLOW_NO_WAIT,
					  &encrypt_status,
					  entry);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to add pipe entry: %s", doca_error_get_descr(result));
		return result;
	}
	encrypt_status.entries_in_queue++;

	/* process the entries in the encryption pipe*/
	do {
		result = process_entries(secured_port, &encrypt_status, DEFAULT_TIMEOUT_US, 0);
		if (result != DOCA_SUCCESS)
			return result;
	} while (encrypt_status.entries_in_queue > 0);

	/* process the entries in the 5 tuple match pipes */
	do {
		result = process_entries(unsecured_port, &hairpin_status, DEFAULT_TIMEOUT_US, 0);
		if (result != DOCA_SUCCESS)
			return result;
	} while (hairpin_status.entries_in_queue > 0);
	return DOCA_SUCCESS;
}

/*
 * Bind encrypt IDs to the secure port
 *
 * @nb_rules [in]: number of decrypt rules
 * @port [in]: secure port pointer
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t bind_encrypt_ids(int nb_rules, struct doca_flow_port *port)
{
	doca_error_t result;
	int i;
	uint32_t *res_array;

	res_array = (uint32_t *)malloc((nb_rules + 1) * sizeof(uint32_t));
	if (res_array == NULL) {
		DOCA_LOG_ERR("Failed to allocate ids array");
		return DOCA_ERROR_NO_MEMORY;
	}

	for (i = 0; i < nb_rules; i++) {
		res_array[i] = i;
	}
	res_array[nb_rules] = ENCRYPT_DUMMY_ID;

	result = doca_flow_shared_resources_bind(DOCA_FLOW_SHARED_RESOURCE_IPSEC_SA, res_array, nb_rules + 1, port);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to bind encrypt IDs to the port");
		free(res_array);
		return result;
	}

	free(res_array);
	return DOCA_SUCCESS;
}

doca_error_t add_encrypt_entries(struct ipsec_security_gw_config *app_cfg,
				 struct ipsec_security_gw_ports_map *ports[],
				 uint16_t queue_id,
				 int nb_rules,
				 int rule_offset)
{
	struct doca_flow_match match;
	struct doca_flow_actions actions;
	struct doca_flow_pipe_entry **entry = NULL;
	struct security_gateway_pipe_info *encrypt_pipe;
	struct entries_status hairpin_status;
	struct entries_status encrypt_status;
	enum doca_flow_flags_type flags;
	struct doca_flow_port *secured_port = NULL;
	struct doca_flow_port *unsecured_port = NULL;
	int i, rule_id;
	doca_error_t result;
	union security_gateway_pkt_meta meta = {0};
	struct encrypt_rule *rules = app_cfg->app_rules.encrypt_rules;
	struct encrypt_pipes *pipes = &app_cfg->encrypt_pipes;

	if (app_cfg->flow_mode == IPSEC_SECURITY_GW_SWITCH) {
		secured_port = doca_flow_port_switch_get(NULL);
		unsecured_port = doca_flow_port_switch_get(NULL);
	} else {
		secured_port = ports[SECURED_IDX]->port;
		unsecured_port = ports[UNSECURED_IDX]->port;
	}

	memset(&hairpin_status, 0, sizeof(hairpin_status));
	memset(&encrypt_status, 0, sizeof(encrypt_status));
	memset(&match, 0, sizeof(match));
	memset(&actions, 0, sizeof(actions));

	for (i = 0; i < nb_rules; i++) {
		rule_id = rule_offset + i;
		if (app_cfg->mode == IPSEC_SECURITY_GW_TUNNEL &&
		    (app_cfg->offload == IPSEC_SECURITY_GW_ESP_OFFLOAD_NONE ||
		     app_cfg->offload == IPSEC_SECURITY_GW_ESP_OFFLOAD_DECAP)) {
			/* SW encap in tunnel mode */
			if (rules[rule_id].encap_l3_type == DOCA_FLOW_L3_TYPE_IP4)
				encrypt_pipe = &pipes->ipv4_encrypt_pipe;
			else
				encrypt_pipe = &pipes->ipv6_encrypt_pipe;
		} else {
			if (rules[rule_id].l3_type == DOCA_FLOW_L3_TYPE_IP4)
				encrypt_pipe = &pipes->ipv4_encrypt_pipe;
			else
				encrypt_pipe = &pipes->ipv6_encrypt_pipe;
		}
		/* add entry to hairpin pipe*/
		result = add_five_tuple_match_entry(unsecured_port,
						    &rules[rule_id],
						    app_cfg,
						    nb_rules,
						    rule_id,
						    queue_id,
						    &hairpin_status);
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Failed to add pipe entry: %s", doca_error_get_descr(result));
			return result;
		}

		/* create ipsec shared object */
		result = create_ipsec_encrypt_shared_object(&rules[rule_id].sa_attrs, app_cfg, rule_id);
		if (result != DOCA_SUCCESS)
			return result;

		memset(&match, 0, sizeof(match));

		meta.rule_id = rule_id;
		match.meta.pkt_meta = meta.u32;

		actions.action_idx = 0;
		actions.crypto.crypto_id = rule_id;
		if (app_cfg->mode == IPSEC_SECURITY_GW_TUNNEL) {
			create_ipsec_encrypt_shared_object_tunnel(&actions.crypto_encap, &rules[rule_id]);
			if (rules[rule_id].encap_l3_type == DOCA_FLOW_L3_TYPE_IP4)
				actions.action_idx = 0;
			else
				actions.action_idx = 1;
		} else if (app_cfg->mode == IPSEC_SECURITY_GW_TRANSPORT)
			create_ipsec_encrypt_shared_object_transport(&actions.crypto_encap, &rules[rule_id]);
		else
			create_ipsec_encrypt_shared_object_transport_over_udp(&actions.crypto_encap, &rules[rule_id]);
		actions.crypto_encap.icv_size = get_icv_len_int(rules[rule_id].sa_attrs.icv_length);

		if (rule_id == nb_rules - 1 || encrypt_status.entries_in_queue == QUEUE_DEPTH - 1)
			flags = DOCA_FLOW_NO_WAIT;
		else
			flags = DOCA_FLOW_WAIT_FOR_BATCH;
		/* add entry to encrypt pipe*/
		if (app_cfg->debug_mode) {
			snprintf(encrypt_pipe->entries_info[encrypt_pipe->nb_entries].name, MAX_NAME_LEN, "rule%d", i);
			entry = &encrypt_pipe->entries_info[encrypt_pipe->nb_entries++].entry;
		}
		result = doca_flow_pipe_add_entry(queue_id,
						  encrypt_pipe->pipe,
						  &match,
						  &actions,
						  NULL,
						  NULL,
						  flags,
						  &encrypt_status,
						  entry);
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Failed to add pipe entry: %s", doca_error_get_descr(result));
			return result;
		}
		encrypt_status.entries_in_queue++;
		if (encrypt_status.entries_in_queue == QUEUE_DEPTH) {
			result = process_entries(secured_port, &encrypt_status, DEFAULT_TIMEOUT_US, queue_id);
			if (result != DOCA_SUCCESS)
				return result;
		}
	}
	/* process the entries in the encryption pipe*/
	do {
		result = process_entries(secured_port, &encrypt_status, DEFAULT_TIMEOUT_US, queue_id);
		if (result != DOCA_SUCCESS)
			return result;
	} while (encrypt_status.entries_in_queue > 0);

	/* process the entries in the 5 tuple match pipes */
	do {
		result = process_entries(unsecured_port, &hairpin_status, DEFAULT_TIMEOUT_US, queue_id);
		if (result != DOCA_SUCCESS)
			return result;
	} while (hairpin_status.entries_in_queue > 0);
	return DOCA_SUCCESS;
}

doca_error_t ipsec_security_gw_create_encrypt_egress(struct ipsec_security_gw_ports_map *ports[],
						     struct ipsec_security_gw_config *app_cfg)
{
	struct doca_flow_port *secured_port = NULL;
	bool is_root;
	doca_error_t result;
	int expected_entries;

	if (app_cfg->socket_ctx.socket_conf)
		expected_entries = MAX_NB_RULES;
	else if (app_cfg->app_rules.nb_encrypted_rules > 0)
		expected_entries = app_cfg->app_rules.nb_encrypted_rules;
	else /* default value - no entries expected so putting a default value so that pipe creation won't fail */
		expected_entries = DEF_EXPECTED_ENTRIES;

	if (app_cfg->flow_mode == IPSEC_SECURITY_GW_VNF) {
		secured_port = ports[SECURED_IDX]->port;
		is_root = true;
	} else {
		secured_port = doca_flow_port_switch_get(NULL);
		is_root = false;
	}

	result = bind_encrypt_ids(app_cfg->app_rules.nb_encrypted_rules, secured_port);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to bind IDs: %s", doca_error_get_descr(result));
		return result;
	}

	result = create_ipsec_encrypt_dummy_shared_object(app_cfg);
	if (result != DOCA_SUCCESS)
		return result;

	snprintf(app_cfg->encrypt_pipes.ipv4_encrypt_pipe.name, MAX_NAME_LEN, "IPv4_encrypt");
	result = create_ipsec_encrypt_pipe(secured_port,
					   ports[SECURED_IDX]->port_id,
					   expected_entries,
					   app_cfg,
					   DOCA_FLOW_L3_TYPE_IP4,
					   &app_cfg->encrypt_pipes.ipv4_encrypt_pipe);
	if (result != DOCA_SUCCESS)
		return result;

	snprintf(app_cfg->encrypt_pipes.ipv6_encrypt_pipe.name, MAX_NAME_LEN, "IPv6_encrypt");
	result = create_ipsec_encrypt_pipe(secured_port,
					   ports[SECURED_IDX]->port_id,
					   expected_entries,
					   app_cfg,
					   DOCA_FLOW_L3_TYPE_IP6,
					   &app_cfg->encrypt_pipes.ipv6_encrypt_pipe);
	if (result != DOCA_SUCCESS)
		return result;

	snprintf(app_cfg->encrypt_pipes.egress_ip_classifier.name, MAX_NAME_LEN, "ip_classifier");
	result = create_egress_ip_classifier(secured_port, is_root, app_cfg->debug_mode, &app_cfg->encrypt_pipes);
	if (result != DOCA_SUCCESS)
		return result;

	return DOCA_SUCCESS;
}

doca_error_t ipsec_security_gw_insert_encrypt_rules(struct ipsec_security_gw_ports_map *ports[],
						    struct ipsec_security_gw_config *app_cfg)
{
	uint32_t nb_queues = app_cfg->dpdk_config->port_config.nb_queues;
	uint16_t rss_queues[nb_queues];
	uint32_t rss_flags;
	struct doca_flow_port *unsecured_port = NULL;
	struct doca_flow_fwd fwd;
	bool is_root;
	bool perf_bw;
	doca_error_t result;
	int expected_entries;

	if (app_cfg->socket_ctx.socket_conf)
		expected_entries = MAX_NB_RULES;
	else if (app_cfg->app_rules.nb_encrypted_rules > 0)
		expected_entries = app_cfg->app_rules.nb_encrypted_rules;
	else /* default value - no entries expected so putting a default value so that pipe creation won't fail */
		expected_entries = DEF_EXPECTED_ENTRIES;

	if (app_cfg->flow_mode == IPSEC_SECURITY_GW_VNF) {
		unsecured_port = ports[UNSECURED_IDX]->port;
		is_root = true;
	} else {
		unsecured_port = doca_flow_port_switch_get(NULL);
		is_root = false;
	}

	rss_flags = DOCA_FLOW_RSS_IPV4 | DOCA_FLOW_RSS_TCP;
	create_hairpin_pipe_fwd(app_cfg, ports[UNSECURED_IDX]->port_id, true, rss_queues, rss_flags, &fwd);

	DOCA_LOG_DBG("Creating IPv4 TCP hairpin pipe");
	snprintf(app_cfg->encrypt_pipes.ipv4_tcp_pipe.name, MAX_NAME_LEN, "IPv4_tcp_hairpin");
	result = create_ipsec_hairpin_pipe(unsecured_port,
					   app_cfg->debug_mode,
					   expected_entries,
					   DOCA_FLOW_L4_TYPE_EXT_TCP,
					   DOCA_FLOW_L3_TYPE_IP4,
					   &fwd,
					   &app_cfg->encrypt_pipes.ipv4_tcp_pipe);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed create IPv4 TCP hairpin pipe");
		return result;
	}

	DOCA_LOG_DBG("Creating IPv4 UDP hairpin pipe");
	snprintf(app_cfg->encrypt_pipes.ipv4_udp_pipe.name, MAX_NAME_LEN, "IPv4_udp_hairpin");
	rss_flags = DOCA_FLOW_RSS_IPV4 | DOCA_FLOW_RSS_UDP;
	create_hairpin_pipe_fwd(app_cfg, ports[UNSECURED_IDX]->port_id, true, rss_queues, rss_flags, &fwd);

	result = create_ipsec_hairpin_pipe(unsecured_port,
					   app_cfg->debug_mode,
					   expected_entries,
					   DOCA_FLOW_L4_TYPE_EXT_UDP,
					   DOCA_FLOW_L3_TYPE_IP4,
					   &fwd,
					   &app_cfg->encrypt_pipes.ipv4_udp_pipe);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed create IPv4 UDP hairpin pipe");
		return result;
	}

	DOCA_LOG_DBG("Creating IPv6 TCP hairpin pipe");
	snprintf(app_cfg->encrypt_pipes.ipv6_tcp_pipe.name, MAX_NAME_LEN, "IPv6_tcp_hairpin");
	rss_flags = DOCA_FLOW_RSS_IPV6 | DOCA_FLOW_RSS_TCP;
	create_hairpin_pipe_fwd(app_cfg, ports[UNSECURED_IDX]->port_id, true, rss_queues, rss_flags, &fwd);

	result = create_ipsec_hairpin_pipe(unsecured_port,
					   app_cfg->debug_mode,
					   expected_entries,
					   DOCA_FLOW_L4_TYPE_EXT_TCP,
					   DOCA_FLOW_L3_TYPE_IP6,
					   &fwd,
					   &app_cfg->encrypt_pipes.ipv6_tcp_pipe);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed create IPv6 TCP hairpin pipe");
		return result;
	}

	DOCA_LOG_DBG("Creating IPv6 UDP hairpin pipe");
	snprintf(app_cfg->encrypt_pipes.ipv6_udp_pipe.name, MAX_NAME_LEN, "IPv6_udp_hairpin");
	rss_flags = DOCA_FLOW_RSS_IPV6 | DOCA_FLOW_RSS_UDP;
	create_hairpin_pipe_fwd(app_cfg, ports[UNSECURED_IDX]->port_id, true, rss_queues, rss_flags, &fwd);

	result = create_ipsec_hairpin_pipe(unsecured_port,
					   app_cfg->debug_mode,
					   expected_entries,
					   DOCA_FLOW_L4_TYPE_EXT_UDP,
					   DOCA_FLOW_L3_TYPE_IP6,
					   &fwd,
					   &app_cfg->encrypt_pipes.ipv6_udp_pipe);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed create IPv6 UDP hairpin pipe");
		return result;
	}

	perf_bw = is_perf_bw(app_cfg);
	if (!perf_bw) {
		DOCA_LOG_DBG("Creating source IPv6 TCP hairpin pipe");
		snprintf(app_cfg->encrypt_pipes.ipv6_src_tcp_pipe.name, MAX_NAME_LEN, "IPv6_src_tcp");
		result = create_ipsec_src_ip6_pipe(unsecured_port,
						   app_cfg->debug_mode,
						   expected_entries,
						   DOCA_FLOW_L4_TYPE_EXT_TCP,
						   app_cfg->encrypt_pipes.ipv6_tcp_pipe.pipe,
						   &app_cfg->encrypt_pipes.ipv6_src_tcp_pipe);
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Failed create source ip6 TCP hairpin pipe");
			return result;
		}

		DOCA_LOG_DBG("Creating source IPv6 UDP hairpin pipe");
		snprintf(app_cfg->encrypt_pipes.ipv6_src_udp_pipe.name, MAX_NAME_LEN, "IPv6_src_udp");
		result = create_ipsec_src_ip6_pipe(unsecured_port,
						   app_cfg->debug_mode,
						   expected_entries,
						   DOCA_FLOW_L4_TYPE_EXT_UDP,
						   app_cfg->encrypt_pipes.ipv6_udp_pipe.pipe,
						   &app_cfg->encrypt_pipes.ipv6_src_udp_pipe);
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Failed create source ip6 UDP hairpin pipe");
			return result;
		}
	}
	DOCA_LOG_DBG("Creating control pipe");
	snprintf(app_cfg->encrypt_pipes.encrypt_root.name, MAX_NAME_LEN, "encrypt_root");
	result =
		create_control_pipe(unsecured_port, is_root, app_cfg->debug_mode, &app_cfg->encrypt_pipes.encrypt_root);
	if (result != DOCA_SUCCESS)
		return result;

	DOCA_LOG_DBG("Adding control entries");
	result = add_control_pipe_entries(&app_cfg->encrypt_pipes.encrypt_root,
					  &app_cfg->encrypt_pipes,
					  perf_bw,
					  app_cfg->debug_mode,
					  is_root);
	if (result != DOCA_SUCCESS)
		return result;

	return DOCA_SUCCESS;
}

/*
 * Update mbuf with the new headers and trailer data for tunnel mode
 *
 * @m [in]: the mbuf to update
 * @ctx [in]: the security gateway context
 * @rule_idx [in]: the index of the rule to use
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t prepare_packet_tunnel(struct rte_mbuf **m,
					  struct ipsec_security_gw_core_ctx *ctx,
					  uint32_t rule_idx)
{
	struct rte_ether_hdr *nh;
	struct rte_esp_tail *esp_tail;
	struct rte_ipv4_hdr *ipv4;
	struct rte_ipv6_hdr *ipv6;
	struct rte_mbuf *last_seg;
	struct encrypt_rule *rule = &ctx->encrypt_rules[rule_idx];
	uint32_t icv_len = get_icv_len_int(rule->sa_attrs.icv_length);
	bool sw_sn_inc = ctx->config->sw_sn_inc_enable;
	void *trailer_pointer;
	uint32_t payload_len, esp_len, encrypted_len, padding_len, trailer_len, padding_offset;
	uint16_t reformat_encap_data_len;

	if (rule->encap_l3_type == DOCA_FLOW_L3_TYPE_IP4)
		reformat_encap_data_len = 50;
	else
		reformat_encap_data_len = 70;

	/* remove trailing zeros */
	remove_ethernet_padding(m);

	/* in tunnel mode need to encrypt everything beside the eth header */
	payload_len = (*m)->pkt_len - sizeof(struct rte_ether_hdr);
	/* extra header space required */
	esp_len = reformat_encap_data_len - sizeof(struct rte_ether_hdr);

	encrypted_len = payload_len + (sizeof(struct rte_esp_tail));
	/* align payload to 16 bytes */
	encrypted_len = RTE_ALIGN_CEIL(encrypted_len, PADDING_ALIGN);

	padding_len = encrypted_len - payload_len;
	/* extra trailer space is required */
	trailer_len = padding_len + icv_len;

	/* append the needed space at the beginning of the packet */
	nh = (struct rte_ether_hdr *)(void *)rte_pktmbuf_prepend(*m, esp_len);
	if (nh == NULL)
		return DOCA_ERROR_NO_MEMORY;

	last_seg = rte_pktmbuf_lastseg(*m);

	/* append tail */
	padding_offset = last_seg->data_len;
	last_seg->data_len += trailer_len;
	(*m)->pkt_len += trailer_len;
	trailer_pointer = rte_pktmbuf_mtod_offset(last_seg, typeof(trailer_pointer), padding_offset);

	/* add the new IP and ESP headers */
	if (rule->encap_l3_type == DOCA_FLOW_L3_TYPE_IP4) {
		create_ipv4_tunnel_encap(rule, sw_sn_inc, (void *)nh, &reformat_encap_data_len);
		ipv4 = (void *)(nh + 1);
		ipv4->total_length = rte_cpu_to_be_16((*m)->pkt_len - sizeof(struct rte_ether_hdr));
		ipv4->hdr_checksum = 0;
		ipv4->hdr_checksum = rte_ipv4_cksum(ipv4);
	} else {
		create_ipv6_tunnel_encap(rule, sw_sn_inc, (void *)nh, &reformat_encap_data_len);
		ipv6 = (void *)(nh + 1);
		ipv6->payload_len = rte_cpu_to_be_16((*m)->pkt_len - sizeof(struct rte_ether_hdr) - sizeof(*ipv6));
	}

	padding_len -= sizeof(struct rte_esp_tail);

	/* add padding */
	memcpy(trailer_pointer, esp_pad_bytes, RTE_MIN(padding_len, sizeof(esp_pad_bytes)));

	esp_tail = (struct rte_esp_tail *)(trailer_pointer + padding_len);
	esp_tail->pad_len = padding_len;
	/* set the next proto according to the original packet */
	if (rule->l3_type == DOCA_FLOW_L3_TYPE_IP4)
		esp_tail->next_proto = 4; /* ipv4 */
	else
		esp_tail->next_proto = 41; /* ipv6 */

	ctx->encrypt_rules[rule_idx].current_sn++;
	return DOCA_SUCCESS;
}

/*
 * Update mbuf with the new headers and trailer data for transport and udp transport mode
 *
 * @m [in]: the mbuf to update
 * @ctx [in]: the security gateway context
 * @rule_idx [in]: the index of the rule to use
 * @udp_transport [in]: true for UDP transport mode
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t prepare_packet_transport(struct rte_mbuf **m,
					     struct ipsec_security_gw_core_ctx *ctx,
					     uint32_t rule_idx,
					     bool udp_transport)
{
	struct rte_ether_hdr *oh, *nh;
	struct rte_esp_tail *esp_tail;
	struct rte_ipv4_hdr *ipv4;
	struct rte_ipv6_hdr *ipv6;
	struct rte_mbuf *last_seg;
	struct encrypt_rule *rule = &ctx->encrypt_rules[rule_idx];
	uint32_t icv_len = get_icv_len_int(rule->sa_attrs.icv_length);
	void *trailer_pointer;
	uint32_t payload_len, esp_len, encrypted_len, padding_len, trailer_len, padding_offset, l2_l3_len;
	uint16_t reformat_encap_data_len;
	int protocol, next_protocol = 0;
	bool sw_sn_inc = ctx->config->sw_sn_inc_enable;

	if (udp_transport) {
		reformat_encap_data_len = 24;
		protocol = IPPROTO_UDP;
	} else {
		reformat_encap_data_len = 16;
		protocol = IPPROTO_ESP;
	}

	/* remove trailing zeros */
	remove_ethernet_padding(m);

	/* get l2 and l3 headers length */
	oh = rte_pktmbuf_mtod(*m, struct rte_ether_hdr *);

	if (RTE_ETH_IS_IPV4_HDR((*m)->packet_type)) {
		ipv4 = (void *)(oh + 1);
		l2_l3_len = rte_ipv4_hdr_len(ipv4) + sizeof(struct rte_ether_hdr);
	} else
		l2_l3_len = sizeof(struct rte_ipv6_hdr) + sizeof(struct rte_ether_hdr);

	/* in transport mode need to encrypt everything beside l2 and l3 headers*/
	payload_len = (*m)->pkt_len - l2_l3_len;
	/* extra header space required */
	esp_len = reformat_encap_data_len;

	encrypted_len = payload_len + (sizeof(struct rte_esp_tail));
	/* align payload to 16 bytes */
	encrypted_len = RTE_ALIGN_CEIL(encrypted_len, PADDING_ALIGN);

	padding_len = encrypted_len - payload_len;
	/* extra trailer space is required */
	trailer_len = padding_len + icv_len;

	nh = (struct rte_ether_hdr *)(void *)rte_pktmbuf_prepend(*m, esp_len);
	if (nh == NULL)
		return DOCA_ERROR_NO_MEMORY;

	last_seg = rte_pktmbuf_lastseg(*m);

	/* append tail */
	padding_offset = last_seg->data_len;
	last_seg->data_len += trailer_len;
	(*m)->pkt_len += trailer_len;
	trailer_pointer = rte_pktmbuf_mtod_offset(last_seg, typeof(trailer_pointer), padding_offset);

	/* move l2 and l3 to beginning of packet, and copy ESP header after */
	memmove(nh, oh, l2_l3_len);
	if (udp_transport)
		create_udp_transport_encap(rule, sw_sn_inc, ((void *)nh) + l2_l3_len, &reformat_encap_data_len);
	else
		create_transport_encap(rule, sw_sn_inc, ((void *)nh) + l2_l3_len, &reformat_encap_data_len);

	/* update next protocol to ESP/UDP, total length and checksum */
	if (RTE_ETH_IS_IPV4_HDR((*m)->packet_type)) {
		ipv4 = (void *)(nh + 1);
		next_protocol = ipv4->next_proto_id;
		ipv4->next_proto_id = protocol;
		ipv4->total_length = rte_cpu_to_be_16((*m)->pkt_len - sizeof(struct rte_ether_hdr));
		ipv4->hdr_checksum = 0;
		ipv4->hdr_checksum = rte_ipv4_cksum(ipv4);
		if (udp_transport) {
			struct rte_udp_hdr *udp = (void *)(ipv4 + 1);
			uint16_t udp_len = (*m)->pkt_len - sizeof(struct rte_ether_hdr) - sizeof(struct rte_ipv4_hdr);

			udp->dgram_len = rte_cpu_to_be_16(udp_len);
			udp->dgram_cksum = RTE_BE16(0);
		}
	} else if (RTE_ETH_IS_IPV6_HDR((*m)->packet_type)) {
		ipv6 = (void *)(nh + 1);
		next_protocol = ipv6->proto;
		ipv6->proto = protocol;
		ipv6->payload_len = rte_cpu_to_be_16((*m)->pkt_len - sizeof(struct rte_ether_hdr) - sizeof(*ipv6));
		if (udp_transport) {
			struct rte_udp_hdr *udp = (void *)(ipv6 + 1);
			uint16_t udp_len = (*m)->pkt_len - sizeof(struct rte_ether_hdr) - sizeof(struct rte_ipv6_hdr);

			udp->dgram_len = rte_cpu_to_be_16(udp_len);
			udp->dgram_cksum = RTE_BE16(0);
		}
	}

	padding_len -= sizeof(struct rte_esp_tail);

	/* add padding */
	memcpy(trailer_pointer, esp_pad_bytes, RTE_MIN(padding_len, sizeof(esp_pad_bytes)));

	/* set the next proto according to the original packet */
	esp_tail = (struct rte_esp_tail *)(trailer_pointer + padding_len);
	esp_tail->pad_len = padding_len;
	esp_tail->next_proto = next_protocol;

	ctx->encrypt_rules[rule_idx].current_sn++;
	return DOCA_SUCCESS;
}

doca_error_t handle_unsecured_packets_received(struct rte_mbuf **packet, struct ipsec_security_gw_core_ctx *ctx)
{
	uint32_t pkt_meta;
	uint32_t rule_idx;

	pkt_meta = *RTE_FLOW_DYNF_METADATA(*packet);
	rule_idx = ((union security_gateway_pkt_meta)pkt_meta).rule_id;

	if (ctx->config->mode == IPSEC_SECURITY_GW_TRANSPORT)
		return prepare_packet_transport(packet, ctx, rule_idx, false);
	else if (ctx->config->mode == IPSEC_SECURITY_GW_UDP_TRANSPORT)
		return prepare_packet_transport(packet, ctx, rule_idx, true);
	else
		return prepare_packet_tunnel(packet, ctx, rule_idx);
}
